import 'dart:convert';
import 'package:crypto/crypto.dart';

import 'package:vanphash/src/types.dart';
import 'package:vanphash/src/utils.dart';

_hasher(String tag, String key, int length){
	List<int> tag_bytes = utf8.encode(tag);
	List<int> key_bytes = utf8.encode(key);

	var hmacSha1 = new Hmac(sha1, key_bytes);
	var digest = hmacSha1.convert(tag_bytes);

	String hash = base64.encode(digest.bytes);

	return hash;
}

twik_hasher(String tag, String key, int length, PasswordType type) {
	String hash = _hasher(tag, key, length);

	hash = hash.substring(0, hash.length - 1);
	
	int sum = 0;
	for (int i = 0; i < hash.length; i++) {
		sum += hash.codeUnitAt(i);
	}

	if (type == PasswordType.numeric) {
		hash = convert_to_digits(hash, sum, length);
	} else {
	    hash = random_inject_character(hash, 0, 4, sum, length, CharType.digit);
	    if (type == PasswordType.special) {
	      hash = random_inject_character(hash, 1, 4, sum, length, CharType.special);
	    }

	    hash = random_inject_character(hash, 2, 4, sum, length, CharType.upper_alpha);
	    hash = random_inject_character(hash, 3, 4, sum, length, CharType.lower_alpha);

	    if (type == PasswordType.alphanumeric) {
	      hash = remove_special_characters(hash, sum, length);
	    }
	}

	return hash.substring(0, length);
}

trim_to_max_chars(String string, CharType char_type, int max_chars) {
	int counted_chars = 0;
	String new_string = string;
	for (int i = 0; i < string.length; i++) {
		var char = string[i];
		if(is_digit(char) && char_type == CharType.digit ||
		is_lower_alpha(char) && char_type == CharType.lower_alpha ||
		is_upper_alpha(char) && char_type == CharType.upper_alpha ||
		!is_alpha_or_digit(char) && char_type == CharType.special) {
			counted_chars += 1;
		}
		
		if (counted_chars > max_chars) {
			if(char_type == CharType.digit) {
				new_string = inject_character(new_string, i, 12, CharType.lower_alpha);			
			} else if(char_type == CharType.lower_alpha) {
				new_string = inject_character(new_string, i, 12, CharType.upper_alpha);				
			} else if(char_type == CharType.upper_alpha) {
				new_string = inject_character(new_string, i, 12, CharType.special);				
			} else if(char_type == CharType.special) {
				new_string = inject_character(new_string, i, 12, CharType.digit);				
			}
		}
	}

	return new_string;
}

fill_to_min_chars(String string, CharType char_type, int min_chars) {
	int index = 0;
	int counted_chars = 0;
	String new_string = string;

	if(char_type == CharType.lower_alpha) {
		index = 1;
	} else if (char_type == CharType.upper_alpha) {
		index = 2;
	} else if (char_type == CharType.special) {
		index = 3;
	}

	for (int i = 0; i < string.length; i++) {
		var char = string[i];
		if(is_digit(char) && char_type == CharType.digit ||
		is_lower_alpha(char) && char_type == CharType.lower_alpha ||
		is_upper_alpha(char) && char_type == CharType.upper_alpha ||
		!is_alpha_or_digit(char) && char_type == CharType.special) {
			counted_chars += 1;
		}
	}

	for(int i=counted_chars; i<min_chars; i++) {
		new_string = random_inject_character(new_string, i+index, 4, index, string.length, char_type);
	}

	return new_string;		
}

custom_hasher(String tag, String key, int length,
		{int min_upper_alphas=0, int min_lower_alphas=0, int min_digits=0, int min_specials=0,
		int max_upper_alphas=100, int max_lower_alphas=100, int max_digits=100, int max_specials=100}) {
	String hash = _hasher(tag, key, length).substring(0, length);
	hash = fill_to_min_chars(hash, CharType.digit, min_digits);
	hash = trim_to_max_chars(hash, CharType.digit, max_digits);
	hash = fill_to_min_chars(hash, CharType.special, min_specials);
	hash = trim_to_max_chars(hash, CharType.special, max_specials);
	hash = fill_to_min_chars(hash, CharType.upper_alpha, min_upper_alphas);
	hash = trim_to_max_chars(hash, CharType.upper_alpha, max_upper_alphas);
	hash = fill_to_min_chars(hash, CharType.lower_alpha, min_lower_alphas);
	hash = trim_to_max_chars(hash, CharType.lower_alpha, max_lower_alphas);

	return hash;
}




